@@ -1,47 +1,35 @@
-from typing import List, Optional, Dict, Union
+from socket import timeout
+from typing import Any, List, Optional, Dict, Sequence, Union
 
 import requests
 import logging
+
+from requests import Response
 import aleph_alpha_client
 from aleph_alpha_client.document import Document
+from aleph_alpha_client.embedding import SemanticEmbeddingRequest
+from aleph_alpha_client.explanation import ExplanationRequest
 from aleph_alpha_client.image import ImagePrompt
-from aleph_alpha_client.prompt_item import _to_prompt_item
+from aleph_alpha_client.prompt import _to_prompt_item, _to_serializable_prompt
 
 POOLING_OPTIONS = ["mean", "max", "last_token", "abs_max"]
 
 
-def _to_serializable_prompt(
-    prompt, at_least_one_token=False
-) -> Union[str, List[Dict[str, str]]]:
-    """
-    Validates that a prompt and emits the format suitable for serialization as JSON
-    """
-    if isinstance(prompt, str):
-        if at_least_one_token:
-            if len(prompt) == 0:
-                raise ValueError("prompt must contain at least one character")
-        # Just pass the string through as is.
-        return prompt
-
-    elif isinstance(prompt, list):
-        return [_to_prompt_item(item) for item in prompt]
-
-    raise ValueError(
-        "Invalid prompt. Prompt must either be a string, or a list of valid multimodal propmt items."
-    )
-
-
 class QuotaError(Exception):
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
 
 
 class AlephAlphaClient:
-    def __init__(self, host, token=None, email=None, password=None):
+    def __init__(
+        self, host, token=None, email=None, password=None, request_timeout_seconds=180
+    ):
         if host[-1] != "/":
             host += "/"
         self.host = host
 
+        self.request_timeout_seconds = request_timeout_seconds
+
         # check server version
         expect_release = "1"
         version = self.get_version()
@@ -54,12 +42,12 @@ class AlephAlphaClient:
         self.token = token or self.get_token(email, password)
 
     def get_version(self):
-        response = requests.get(self.host + "version")
+        response = self.get_request(self.host + "version")
         response.raise_for_status()
         return response.text
 
     def get_token(self, email, password):
-        response = requests.post(
+        response = self.post_request(
             self.host + "get_token", json={"email": email, "password": password}
         )
         if response.status_code == 200:
@@ -68,6 +56,14 @@ class AlephAlphaClient:
         else:
             raise ValueError("cannot get token")
 
+    def get_request(self, url, headers=None):
+        return requests.get(url, headers=headers, timeout=self.request_timeout_seconds)
+
+    def post_request(self, url, json, headers=None):
+        return requests.post(
+            url, headers=headers, json=json, timeout=self.request_timeout_seconds
+        )
+
     @property
     def request_headers(self):
         return {
@@ -79,10 +75,10 @@ class AlephAlphaClient:
         """
         Queries all models which are currently available.
         """
-        response = requests.get(
+        response = self.get_request(
             self.host + "models_available", headers=self.request_headers
         )
-        return self._parse_response(response)
+        return self._translate_errors(response)
 
     def tokenize(
         self, model: str, prompt: str, tokens: bool = True, token_ids: bool = True
@@ -96,26 +92,24 @@ class AlephAlphaClient:
             "tokens": tokens,
             "token_ids": token_ids,
         }
-        response = requests.post(
+        response = self.post_request(
             self.host + "tokenize",
             headers=self.request_headers,
             json=payload,
-            timeout=None,
         )
-        return self._parse_response(response)
+        return self._translate_errors(response)
 
     def detokenize(self, model: str, token_ids: List[int]):
         """
         Detokenizes the given tokens.
         """
         payload = {"model": model, "token_ids": token_ids}
-        response = requests.post(
+        response = self.post_request(
             self.host + "detokenize",
             headers=self.request_headers,
             json=payload,
-            timeout=None,
         )
-        return self._parse_response(response)
+        return self._translate_errors(response)
 
     def complete(
         self,
@@ -130,6 +124,8 @@ class AlephAlphaClient:
         frequency_penalty: Optional[float] = 0.0,
         repetition_penalties_include_prompt: Optional[bool] = False,
         use_multiplicative_presence_penalty: Optional[bool] = False,
+        penalty_bias: Optional[str] = None,
+        penalty_exceptions: Optional[List[str]] = None,
         best_of: Optional[int] = None,
         n: Optional[int] = 1,
         logit_bias: Optional[Dict[int, float]] = None,
@@ -212,101 +208,18 @@ class AlephAlphaClient:
         if not isinstance(model, str):
             raise ValueError("model must be a string")
 
-        prompt = _to_serializable_prompt(prompt=prompt)
-
-        if not (maximum_tokens is None or isinstance(maximum_tokens, int)):
-            raise ValueError("maximum_tokens must be an int or None")
         if isinstance(temperature, int):
             temperature = float(temperature)
-        if not (temperature is None or isinstance(temperature, float)):
-            raise ValueError("temperature must be a float or None")
-        if not (top_k is None or isinstance(top_k, int)):
-            raise ValueError("top_k must be a positive int or None")
         if isinstance(top_p, int):
             top_p = float(top_p)
-        if not (top_p is None or isinstance(top_p, float)):
-            raise ValueError("top_p must be a float or None")
         if isinstance(presence_penalty, int):
             presence_penalty = float(presence_penalty)
-        if not (presence_penalty is None or isinstance(presence_penalty, float)):
-            raise ValueError("presence_penalty must be a float or None")
         if isinstance(frequency_penalty, int):
             frequency_penalty = float(frequency_penalty)
-        if not (frequency_penalty is None or isinstance(frequency_penalty, float)):
-            raise ValueError("frequency_penalty must be a float or None")
-        if not (
-            repetition_penalties_include_prompt is None
-            or isinstance(repetition_penalties_include_prompt, bool)
-        ):
-            raise ValueError(
-                "repetition_penalties_include_prompt must be a bool or None"
-            )
-        if not (
-            use_multiplicative_presence_penalty is None
-            or isinstance(use_multiplicative_presence_penalty, bool)
-        ):
-            raise ValueError(
-                "use_multiplicative_presence_penalty must be a bool or None"
-            )
-        if not (best_of is None or isinstance(best_of, int)):
-            raise ValueError("best_of must be an int or None")
-        if not (n is None or isinstance(n, int)):
-            raise ValueError("n must be an int or None")
-        if not (logit_bias is None or isinstance(logit_bias, dict)):
-            raise ValueError("logit_bias must be a dict or None")
-        if logit_bias is not None:
-            for k, v in logit_bias.items():
-                if not isinstance(k, int):
-                    raise ValueError("a key in the logit_bias dict must be an integer")
-                if not isinstance(v, float):
-                    raise ValueError("a value in the logit_bias dict must be a float")
-        if not (log_probs is None or isinstance(log_probs, int)):
-            raise ValueError("log_probs must be an int or None")
-        if not (stop_sequences is None or isinstance(stop_sequences, list)):
-            raise ValueError("stop_sequences must be an list of strings or None")
-        if stop_sequences is not None:
-            for stop_sequence in stop_sequences:
-                if not isinstance(stop_sequence, str):
-                    raise ValueError(
-                        "each item in the stop_sequences list must be a string"
-                    )
-        if not (tokens is None or isinstance(tokens, bool)):
-            raise ValueError("tokens must be a bool or None")
-        if not (
-            disable_optimizations is None or isinstance(disable_optimizations, bool)
-        ):
-            raise ValueError("disable_optimizations must be a bool or None")
-
-        # validate values
-        if maximum_tokens is not None:
-            if maximum_tokens <= 0:
-                raise ValueError("maximum_tokens must be a positive integer")
-        if top_k is not None:
-            if top_k < 0:
-                raise ValueError("top_k must be a positive integer, 0 or None")
-        if temperature is not None:
-            if temperature < 0.0 or temperature > 1.0:
-                raise ValueError("temperature must be a float between 0.0 and 1.0")
-        if top_p is not None:
-            if top_p < 0.0 or top_p > 1.0:
-                raise ValueError("top_p must be a float between 0.0 and 1.0")
-        if n is not None:
-            if n <= 0:
-                raise ValueError("top_k must be a positive integer")
-
-        if best_of is not None:
-            if best_of == n:
-                raise ValueError(
-                    "With best_of equal to n no best completions are choses because only n are computed."
-                )
-            if best_of < n:
-                raise ValueError(
-                    "best_of needs to be bigger than n. The model cannot return more completions (n) than were computed (best_of)."
-                )
 
         payload = {
             "model": model,
-            "prompt": prompt,
+            "prompt": _to_serializable_prompt(prompt=prompt),
             "hosting": hosting,
             "maximum_tokens": maximum_tokens,
             "temperature": temperature,
@@ -314,6 +227,8 @@ class AlephAlphaClient:
             "top_p": top_p,
             "presence_penalty": presence_penalty,
             "frequency_penalty": frequency_penalty,
+            "penalty_bias": penalty_bias,
+            "penalty_exceptions": penalty_exceptions,
             "best_of": best_of,
             "n": n,
             "logit_bias": logit_bias,
@@ -325,13 +240,12 @@ class AlephAlphaClient:
             "disable_optimizations": disable_optimizations,
         }
 
-        response = requests.post(
+        response = self.post_request(
             self.host + "complete",
             headers=self.request_headers,
             json=payload,
-            timeout=None,
         )
-        response_json = self._parse_response(response)
+        response_json = self._translate_errors(response)
         if response_json.get("optimized_prompt") is not None:
             # Return a message to the user that we optimized their prompt
             print(
@@ -341,12 +255,13 @@ class AlephAlphaClient:
 
     def embed(
         self,
-        model,
-        prompt: Union[str, List[Union[str, ImagePrompt]]],
+        model: str,
+        prompt: Union[str, Sequence[Union[str, ImagePrompt]]],
         pooling: List[str],
         layers: List[int],
         hosting: str = "cloud",
         tokens: Optional[bool] = False,
+        type: Optional[str] = None,
     ):
         """
         Embeds a text and returns vectors that can be used for downstream tasks (e.g. semantic similarity) and models (e.g. classifiers).
@@ -379,44 +294,18 @@ class AlephAlphaClient:
 
             tokens (bool, optional, default False)
                 Flag indicating whether the tokenized prompt is to be returned (True) or not (False)
-        """
 
-        if not isinstance(model, str):
-            raise ValueError("model must be a string")
+            type
+                Type of the embedding (e.g. symmetric or asymmetric)
+
+        """
 
         serializable_prompt = _to_serializable_prompt(
             prompt=prompt, at_least_one_token=True
         )
 
-        if not isinstance(layers, list):
-            raise ValueError("layers must be a list")
-
-        if len(layers) == 0:
-            raise ValueError("layers must contain at least one layer")
-
-        for layer in layers:
-            if not isinstance(layer, int):
-                raise ValueError(
-                    "each item in the layers list must be an integer; got " + str(layer)
-                )
-
         if tokens is None:
             tokens = False
-        if not isinstance(tokens, bool):
-            raise ValueError("tokens must be a bool")
-
-        if not pooling is None and not isinstance(pooling, list):
-            raise ValueError("pooling must be None or a list")
-
-        if pooling is not None:
-            for pooling_option in pooling:
-                if pooling_option not in POOLING_OPTIONS:
-                    raise ValueError(
-                        "each item in the pooling list must be either one of "
-                        + str(POOLING_OPTIONS)
-                        + "; got "
-                        + str(pooling_option)
-                    )
 
         payload = {
             "model": model,
@@ -425,11 +314,50 @@ class AlephAlphaClient:
             "layers": layers,
             "tokens": tokens,
             "pooling": pooling,
+            "type": type,
         }
-        response = requests.post(
+        response = self.post_request(
             self.host + "embed", headers=self.request_headers, json=payload
         )
-        return self._parse_response(response)
+        return self._translate_errors(response)
+
+    def semantic_embed(
+        self,
+        model: str,
+        request: SemanticEmbeddingRequest,
+        hosting: str = "cloud",
+    ):
+        """
+        Embeds a text and returns vectors that can be used for downstream tasks (e.g. semantic similarity) and models (e.g. classifiers).
+
+        Parameters:
+            model (str, required):
+                Name of model to use. A model name refers to a model architecture (number of parameters among others). Always the latest version of model is used. The model output contains information as to the model version.
+
+            hosting (str, required):
+                Specifies where the computation will take place. This defaults to "cloud", meaning that it can be
+                executed on any of our servers. An error will be returned if the specified hosting is not available.
+                Check available_models() for available hostings.
+
+            request (SemanticEmbeddingRequest, required)
+                NamedTuple containing all necessary request parameters.
+        """
+
+        serializable_prompt = _to_serializable_prompt(
+            prompt=request.prompt.items, at_least_one_token=True
+        )
+
+        payload: Dict[str, Any] = {
+            "model": model,
+            "hosting": hosting,
+            "prompt": serializable_prompt,
+            "representation": request.representation.value,
+            "compress_to_size": request.compress_to_size,
+        }
+        response = self.post_request(
+            self.host + "semantic_embed", headers=self.request_headers, json=payload
+        )
+        return self._translate_errors(response)
 
     def evaluate(
         self,
@@ -457,27 +385,18 @@ class AlephAlphaClient:
                 The text to be completed. Unconditional completion can be used with an empty string (default). The prompt may contain a zero shot or few shot task.
         """
 
-        if not isinstance(model, str):
-            raise ValueError("model must be a string")
-
         serializable_prompt = _to_serializable_prompt(prompt=prompt)
 
-        if not isinstance(completion_expected, str):
-            raise ValueError("completion_expected must be a string")
-
-        if len(completion_expected) == 0:
-            raise ValueError("completion_expected cannot be empty")
-
         payload = {
             "model": model,
             "prompt": serializable_prompt,
             "hosting": hosting,
             "completion_expected": completion_expected,
         }
-        response = requests.post(
+        response = self.post_request(
             self.host + "evaluate", headers=self.request_headers, json=payload
         )
-        return self._parse_response(response)
+        return self._translate_errors(response)
 
     def qa(
         self,
@@ -489,6 +408,7 @@ class AlephAlphaClient:
         disable_optimizations: bool = False,
         max_answers: int = 0,
         min_score: float = 0.0,
+        hosting: str = "cloud",
     ):
         """
         Answers a question about a prompt.
@@ -526,63 +446,53 @@ class AlephAlphaClient:
 
             min_score (float, default 0.0):
                 The lower limit of minimum score for every answer.
+
+            hosting (str, default "cloud"):
+                Specifies where the computation will take place. This defaults to "cloud", meaning that it can be
+                executed on any of our servers. An error will be returned if the specified hosting is not available.
+                Check available_models() for available hostings.
         """
 
-        # validate data types
-        if not isinstance(model, str):
-            raise ValueError("model must be a string")
-
-        if not isinstance(query, str):
-            raise ValueError("query must be a string")
-
-        if not isinstance(documents, list):
-            raise ValueError(
-                "documents must be a list where all elements are of the type Document"
-            )
-
-        documents = [document._to_serializable_document() for document in documents]
-
-        if not isinstance(maximum_tokens, int):
-            raise ValueError("maximum_tokens must be an int")
-
-        if not isinstance(max_chunk_size, int):
-            raise ValueError("max_chunk_size must be an int")
-
-        if not isinstance(max_answers, int):
-            raise ValueError("max_answers must be an int")
-
-        if not isinstance(min_score, float):
-            raise ValueError("min_score must be a float")
-
-        if not isinstance(disable_optimizations, bool):
-            raise ValueError("disable_optimizations must be a bool")
-
-        # validate values
-        if maximum_tokens <= 0:
-            raise ValueError("maximum_tokens must be a positive integer")
-
         payload = {
             "model": model,
             "query": query,
-            "documents": documents,
+            "documents": [
+                document._to_serializable_document() for document in documents
+            ],
             "maximum_tokens": maximum_tokens,
             "max_answers": max_answers,
             "min_score": min_score,
             "max_chunk_size": max_chunk_size,
             "disable_optimizations": disable_optimizations,
+            "hosting": hosting,
         }
 
-        response = requests.post(
+        response = self.post_request(
             self.host + "qa",
             headers=self.request_headers,
             json=payload,
-            timeout=None,
         )
-        response_json = self._parse_response(response)
+        response_json = self._translate_errors(response)
         return response_json
 
+    def _explain(
+        self, model: str, request: ExplanationRequest, hosting: Optional[str] = None
+    ):
+        body = {
+            "model": model,
+            "prompt": [_to_prompt_item(item) for item in request.prompt.items],
+            "target": request.target,
+            "suppression_factor": request.suppression_factor,
+            "directional": request.directional,
+            "conceptual_suppression_threshold": request.conceptual_suppression_threshold,
+        }
+        response = self.post_request(
+            f"{self.host}explain", headers=self.request_headers, json=body
+        )
+        return self._translate_errors(response)
+
     @staticmethod
-    def _parse_response(response):
+    def _translate_errors(response: Response):
         if response.status_code == 200:
             return response.json()
         else:
